import java.util.ArrayList;
import java.util.List;


/***
 * This class implements a rectangular map.
 * The map is divided into fixed sized squares called map points (MapPoint).
 * Each MapPoint has a unique coordinate.
 * The upper left map point has coordinate (0,0)
 * and the lower right map point (sizeX - 1, sizeY - 1)
 * where the width of the map is sizeY and the height is sizeX in number of squares.
 * (x,y)
 * (0,0)
 * -----------> Y
 * |
 * |
 * |
 * |
 * X
 *
 * Example 5x5
 * (0,0) (0,1) (0,2) (0,3) (0,4)
 * (1,0) (1,1) (1,2) (1,3) (1,4)
 * (2,0) (2,1) (2,2) (2,3) (2,4)
 * (3,0) (3,1) (3,2) (3,3) (3,4)
 * (4,0) (4,1) (4,2) (4,3) (4,4)
 *
 */
public class FactoryMap {
    private MapPoint[][] map;
    private int sizeX;
    private int sizeY;

    /**
     * Initializes a Map without any obstacles
     *
     * @param sizeX the height of the map
     * @param sizeY the width of the map
     */
    public FactoryMap(int sizeX, int sizeY) {
        this.sizeX = sizeX;
        this.sizeY = sizeY;
        createMap();
    }

    /**
     * Initializes a Map with the specified obstacles.
     *
     * @param sizeX the height of the map
     * @param sizeY the width of the map
     */
    public FactoryMap(int sizeX, int sizeY, Obstacle[] obstacles) {
        this(sizeX, sizeY);
        setObstacles(obstacles);
    }

    private void createMap() {
        map = new MapPoint[sizeX][sizeY];
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                map[xCor][yCor] = new MapPoint(xCor, yCor);
            }
        }
    }

    public void setObstacle(Obstacle obst) {
        for (int xCor = obst.x; xCor < obst.x + obst.dx; xCor++) {
            for (int yCor = obst.y; yCor < obst.y + obst.dy; yCor++) {
                map[xCor][yCor].setBlocked();
            }
        }
    }

    public void setObstacles(Obstacle[] obstacles) {
        for (Obstacle obst : obstacles)
            setObstacle(obst);
    }

    /**
     * Returns the MapPoint with the given Coordinates
     *
     * @param x
     * @param y
     * @return
     */
    public MapPoint getPoint(int x, int y) {
        return map[x][y];
    }

    /**
     * Returns all neighbouring MapPoints x where x is not blocked.
     * Bsp (# means blocked):
     * o o #
     * o i o
     * o o o
     * Here all o's would be returned when i is the input to the Method
     *
     * @param point the input MapPoint
     * @return a list of neighbouring points
     */
    public List<MapPoint> getMovableNeighbours(MapPoint point) {
        ArrayList<MapPoint> tmp = new ArrayList<>();
        //TODO
        return tmp;
    }

    public int countNumUnexplored() {
        int count = 0;
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                if (map[xCor][yCor].state == MapPoint.State.UNEXPLORED)
                    count++;
            }
        }
        return count;
    }

    public int getSizeX() {
        return sizeX;
    }

    public int getSizeY() {
        return sizeY;
    }

    public static FactoryMap fromString(String map) {
        int sizeX, sizeY;
        String[] lines = map.split("\n");
        sizeX = lines.length;
        sizeY = lines[0].split(" ").length;
        FactoryMap res = new FactoryMap(sizeX, sizeY);
        for (int xCor = 0; xCor < sizeX; xCor++) {
            String[] tokens = lines[xCor].split(" ");
            for (int yCor = 0; yCor < sizeY; yCor++) {
                switch (tokens[yCor]) {
                    case "#":
                        res.getPoint(xCor, yCor).setBlocked();
                        break;
                    case "o":
                        res.getPoint(xCor, yCor).setExplored();
                        break;
                    case "?":
                        res.getPoint(xCor, yCor).setUnexplored();
                        break;
                    case "S":
                        res.getPoint(xCor, yCor).setStart();
                        break;
                    case "T":
                        res.getPoint(xCor, yCor).setTarget();
                        break;
                    case "X":
                        res.getPoint(xCor, yCor).setPath();
                        break;
                    case "C":
                        res.getPoint(xCor, yCor).setCurrent();
                        break;
                    default:

                }
            }
        }
        return res;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                MapPoint.State state = map[xCor][yCor].state;
                switch (state) {
                    case BLOCKED:
                        sb.append("#");
                        break;
                    case EXPLORED:
                        sb.append("o");
                        break;
                    case UNEXPLORED:
                        sb.append("?");
                        break;
                    case START:
                        sb.append("S");
                        break;
                    case TARGET:
                        sb.append("T");
                        break;
                    case PATH:
                        sb.append("X");
                        break;
                    case CURRENT:
                        sb.append("C");
                        break;
                }
                sb.append(" ");
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    /**
     * Encodes a rectangular area, which is blocked.
     */
    public static class Obstacle {
        int x;
        int y;
        int dx;
        int dy;

        public Obstacle(int x, int y, int dx, int dy) {
            this.x = x;
            this.y = y;
            this.dx = dx;
            this.dy = dy;
        }
    }

    /**
     * Represents a Point on the Map.
     */
    public static class MapPoint {

        enum State {
            BLOCKED,
            EXPLORED,
            UNEXPLORED,
            CURRENT,
            START,
            TARGET,
            PATH
        }

        int x;
        int y;
        public State state;

        private MapPoint(int x, int y) {
            this.x = x;
            this.y = y;
            this.state = State.UNEXPLORED;
        }

        public void setExplored() {
            if (this.state == State.BLOCKED)
                throw new RuntimeException("A blocked node cannot be explored");
            this.state = State.EXPLORED;
        }

        public void setUnexplored() {
            if (this.state == State.BLOCKED)
                throw new RuntimeException("A blocked node cannot be unexplored");
            this.state = State.UNEXPLORED;
        }

        public void setPath() {
            this.state = State.PATH;
        }

        public void setTarget() {
            this.state = State.TARGET;
        }

        public void setCurrent() {
            this.state = State.CURRENT;
        }

        public void setBlocked() {
            this.state = State.BLOCKED;
        }

        public void setStart() {
            this.state = State.START;
        }

        /**
         * Computes the geometric distance between this point (x1, y1) and another point (x2, y2):
         * distance = sqrt((x2-x1)^2 + (y2- y1)^2)
         *
         * @param other the other MapPoint
         * @return the geometric distance to the other point
         */
        public double getDistance(MapPoint other) {
            //TODO
            return 0.0;
        }

        public int getX() {
            return x;
        }

        public int getY() {
            return y;
        }

        public boolean isBlocked() {
            return this.state == State.BLOCKED;
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof MapPoint))
                return false;
            MapPoint point = (MapPoint) obj;
            return point.x == this.x && point.y == this.y;
        }

        @Override
        public int hashCode() {
            int tmp = y + ((x + 1) / 2);
            return x + (tmp * tmp);
        }

    }

}
