import java.util.Comparator;
import java.util.HashMap;


/**
 * Wraps a FactoryMap.MapPoint.
 * Provides useful fields and methods for the A*-Algorithm
 */
public class GraphNode {
    /**
     * The wrapped MapPoint from the FactoryMap
     */
    private FactoryMap.MapPoint point;

    /**
     * The f-Value of the wrapped Point
     */
    private double fValue;

    /**
     * The g-Value of the wrapped Point
     */
    private double gValue;

    /**
     * Points to a previous GraphNode
     */
    private GraphNode previous = null;

    /**
     * DO NOT USE THIS CONSTRUCTOR
     * Use the GraphNodeCreator instead
     */
    private GraphNode(FactoryMap.MapPoint point, double fValue, double gValue) {
        this.point = point;
        this.fValue = fValue;
        this.gValue = gValue;
    }

    /**
     * DO NOT USE THIS CONSTRUCTOR
     * Use the GraphNodeCreator instead
     */
    private GraphNode(FactoryMap.MapPoint point) {
        this(point, Double.MAX_VALUE, Double.MAX_VALUE);
    }

    public double getDistance(GraphNode other) {
        return this.point.getDistance(other.point);
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof GraphNode))
            return false;
        GraphNode node = (GraphNode) obj;
        return this.point.equals(node.point);
    }

    @Override
    public int hashCode() {
        return point.hashCode();
    }

    /**
     * Returns a Comparator<GraphNode> object, which compares GraphNodes according to their f-Value
     */
    public static Comparator<GraphNode> getPriorityComparator() {
        return new Comparator<GraphNode>() {
            @Override
            public int compare(GraphNode o1, GraphNode o2) {
                return Double.compare(o1.fValue, o2.fValue);
            }
        };
    }

    public FactoryMap.MapPoint getPoint() {
        return point;
    }

    public double getfValue() {
        return fValue;
    }

    public double getgValue() {
        return gValue;
    }

    public GraphNode getPrevious() {
        return previous;
    }

    public void setPrevious(GraphNode previous) {
        this.previous = previous;
    }

    public void setfValue(double fValue) {
        this.fValue = fValue;
    }

    public void setgValue(double gValue) {
        this.gValue = gValue;
    }

    /**
     * A factory for GraphNode objects of a specific FactoryMap,
     */
    public static class GraphNodeCreator {

        /**
         * Stores the mappings from Map.MapPoint objects to their wrapped GraphNode objects
         */
        private HashMap<FactoryMap.MapPoint, GraphNode> mapPointToAlgorithmNode = new HashMap<>();

        /**
         * Returns a GraphNode object for the given FactoryMap.MapPoint object.
         * Request for equal FactoryMap.MapPoint objects receive the same GraphNode object.
         *
         * @param point the input FactoryMap.MapPoint
         * @return a GraphNode object
         */
        public GraphNode getAlgorithmNode(FactoryMap.MapPoint point) {
            GraphNode res = mapPointToAlgorithmNode.get(point);
            if (res == null) {
                res = new GraphNode(point);
                mapPointToAlgorithmNode.put(point, res);
            }
            return res;
        }
    }

}
